#include "Interpreter.h"

#include <stdio.h>


const char* cmds[] = {"NOP","PUSH","POP","ADD","SUB","MUL","DIV","MOD","NOT","GRT","PNTR","SWTCH","DUP","ROLL","INN","INC","OUTN","OUTC","EXIT"};

void runProgram(Program& input){
	ProgramCommandList* currentCommand = input.commands;
	int memory[102400]; //100kB: TODO dynamic: double each time there is not enough memory available
	unsigned int stackPosition = 0;
	const unsigned int memSize = 102400;
	unsigned char cond = 0; //store the selected condition

	while (currentCommand->command != EXIT) {
		cond = 0;
		switch (currentCommand->command){
			case NOP:
				break;
			case PUSH:
				memory[stackPosition]=(int)(currentCommand->parameter);
				stackPosition++;
				break;
			case POP:
				if (stackPosition<1)
					break;
				stackPosition--;
				break;
			case ADD:
				if (stackPosition<2)
					break;
				stackPosition--;
				memory[stackPosition-1]=memory[stackPosition-1]+memory[stackPosition];
				break;
			case SUB:
				if (stackPosition<2)
					break;
				stackPosition--;
				memory[stackPosition-1]=memory[stackPosition-1]-memory[stackPosition];
				break;
			case MUL:
				if (stackPosition<2)
					break;
				stackPosition--;
				memory[stackPosition-1]=memory[stackPosition-1]*memory[stackPosition];
				break;
			case DIV:
				if (stackPosition<2)
					break;
				if (memory[stackPosition-1] == 0)
					break; //division by zero
				stackPosition--;
				memory[stackPosition-1]=memory[stackPosition-1]/memory[stackPosition];
				break;
			case MOD:
				if (stackPosition<2)
					break;
				if (memory[stackPosition-1] == 0)
					break; // division by zero
				stackPosition--;
				memory[stackPosition-1]=memory[stackPosition-1]%memory[stackPosition];
				break;
			case NOT:
				if (stackPosition<1)
					break;
				memory[stackPosition-1]=(memory[stackPosition-1]==0?1:0);
				break;
			case GRT:
				if (stackPosition<2)
					break;
				stackPosition--;
				memory[stackPosition-1]=(memory[stackPosition-1]>memory[stackPosition]?1:0);
				break;
			case PNTR:
				if (stackPosition<1)
					break;
				stackPosition--;
				if (memory[stackPosition] >= 0)
					cond = (memory[stackPosition]%4);
				else
					cond = (4-((-memory[stackPosition])%4))%4;
				break;
			case SWTCH:
				if (stackPosition<1)
					break;
				stackPosition--;
				if (memory[stackPosition] >= 0)
					cond = (memory[stackPosition]%2);
				else
					cond = (-memory[stackPosition])%2;
				break;
			case DUP:
				memory[stackPosition]=memory[stackPosition-1];
				stackPosition++;
				break;
			case ROLL:
				if (stackPosition<2)
					break;
				if (memory[stackPosition-2]<=0)
					break; //negative or zero depth
				if ((unsigned int)(memory[stackPosition-2])>stackPosition-2)
					break; //not enough elements at the stack
				{
					int depth = memory[stackPosition-2];
					int offset = stackPosition-2-depth;
					int roll = memory[stackPosition-1];
					int* tmp = new int[memory[stackPosition-2]];
					if (tmp==0){
						//out of memory!
						// TODO: clean up heap!
						return;
					}
					for (int i=0;i<depth;i++){
						tmp[i] = memory[offset+i];
					}
					// now roll by copying back the values...
					for (int i=0;i<depth;i++){
						int pos = i-roll;
						if (pos<0){
							pos = (memory[stackPosition-2])-((-pos)%(memory[stackPosition-2]));
						}
						memory[offset+i] = tmp[pos%depth];
					}
					stackPosition-=2;
					delete tmp;
				}
				break;
			case OUTN:
				if (stackPosition<1)
					break;
				stackPosition--;
				printf("%d",memory[stackPosition]);
				break;
			case OUTC:
				if (stackPosition<1)
					break;
				stackPosition--;
				printf("%c",(char)(memory[stackPosition]));
				break;
			case INN:
				// TODO: read in arbitrary characters and parse them to int "by hand".
				memory[stackPosition]=0;
				scanf("%d",memory+stackPosition);
				stackPosition++;
				break;
			case INC:
				{
					unsigned char in=0;
					scanf("%c",&in);
					memory[stackPosition]=in;
					stackPosition++;
				}
				break;
			default:
				// internal error!
				// TODO: print out error message!
				// TODO: clean up heap if used...
				//printf("Invalid CMD: %d\n",currentCommand->command);
				return;
				
		}

		currentCommand = currentCommand->cond[cond];
		if (currentCommand == 0){
			//printf("ERROR: nextCMD is 0\n");
			return; // TODO: print out error message! //TODO: clean up heap if used...
		}
	}

	// EXIT-Cmd reached.
	//TODO: clean up heap if used...
	return;
}

