#include "LZW.h"
#include <stdio.h>


LZW::LZW(unsigned char bitdepth){
	this->bitdepth = bitdepth;
	this->currCodeLen = bitdepth+1;
	this->lastCodeTableEntry = (1<<bitdepth)+1;
	for (unsigned short i=0;i<=this->lastCodeTableEntry;i++){
		codeTable[i].code = new unsigned char;
		*(codeTable[i].code) = (unsigned char)i;
		codeTable[i].codeLen = 1;
	}
	lastCodedSequence.code = 0;
	lastCodedSequence.codeLen = 0;
	tempBitsBitCount = 0;
	this->finished = false;
}

bool LZW::isFinished(){
	return this->finished;
}

void LZW::reset(){
	this->currCodeLen = this->bitdepth+1;
	this->lastCodeTableEntry = (1<<this->bitdepth)+1; // EOF
	lastCodedSequence.code = 0;
	lastCodedSequence.codeLen = 0;
	tempBitsBitCount = 0;
	this->finished = false;
}

void LZW::decodeByte(unsigned char input, char*& output, unsigned short& outputLen){
	output = 0;
	outputLen = 0;

	if (finished)
		return;

	this->tempBits &= ((1<<this->tempBitsBitCount)-1);
	this->tempBits |= (((unsigned int)input)<<this->tempBitsBitCount);	
	this->tempBitsBitCount += 8;

	while (this->currCodeLen<=this->tempBitsBitCount){
		unsigned short readCTI = this->tempBits & ((1<<this->currCodeLen)-1);
		this->tempBits>>=this->currCodeLen;
		this->tempBitsBitCount -= this->currCodeLen;
		if (readCTI == (1<<bitdepth)){
			this->currCodeLen = this->bitdepth+1;
			this->lastCodeTableEntry = (1<<this->bitdepth)+1;
			this->lastCodedSequence.codeLen = 0;
			if (this->lastCodedSequence.code != 0)
				delete this->lastCodedSequence.code;
			this->lastCodedSequence.code = 0;
			continue;
		}
		if (readCTI == (1<<this->bitdepth)+1){
			//END OF LZW DATA
			this->finished = true;
			return;
		}
		if (readCTI > this->lastCodeTableEntry+1){
//			if (verbose)
//				printf("Warning: Invalid code table index in LZW code.\n");
			continue;
		}
		//schreibe alten code ins wörterbuch, falls er länger als 0 ist
		if (this->lastCodedSequence.codeLen > 0 && this->lastCodeTableEntry < 4095){
			this->codeTable[++this->lastCodeTableEntry].codeLen = this->lastCodedSequence.codeLen+1;
			this->codeTable[this->lastCodeTableEntry].code = new unsigned char[codeTable[this->lastCodeTableEntry].codeLen];
			for (unsigned short i=0;i<this->lastCodedSequence.codeLen;i++){
				this->codeTable[this->lastCodeTableEntry].code[i] = this->lastCodedSequence.code[i];
			}
			//hänge erstes zeichen des gelesenen codes an den woerterbucheintrag an
			this->codeTable[this->lastCodeTableEntry].code[lastCodedSequence.codeLen] = this->codeTable[readCTI].code[0];
		}

		unsigned short oldOutputLen = outputLen;
		char* oldOutput = output;
		outputLen += this->codeTable[readCTI].codeLen;
		this->lastCodedSequence.codeLen = this->codeTable[readCTI].codeLen;
		if (this->lastCodedSequence.code != 0)
			delete this->lastCodedSequence.code;
		this->lastCodedSequence.code = new unsigned char[this->lastCodedSequence.codeLen];

		output = new char[outputLen+1];
		output[outputLen] = 0;
		for (unsigned short i=0;i<oldOutputLen;i++){
			output[i] = oldOutput[i];
		}
		if (oldOutput != 0)
			delete oldOutput;
		for (unsigned short i=0;i<this->codeTable[readCTI].codeLen;i++){
			this->lastCodedSequence.code[i] = this->codeTable[readCTI].code[i];
			output[oldOutputLen+i] = this->codeTable[readCTI].code[i];
		}
		if (this->lastCodeTableEntry >= (1<<this->currCodeLen)-1){
			this->currCodeLen++;
			if (this->currCodeLen>12){
				this->currCodeLen = 12;
			}
		}
	}
	return;

}

LZW::~LZW(){
	//TODO
}

