#include "PietImage2Program.h"

int pietimage2program(PietImage input, Program& output, bool spaceAfterNumOut, bool printQuestionMarkOnIn){

	Region** regionmap; //Map of regions: A Pointer to a corresponding Region for every codel.
	regionmap = new Region*[input.size[0]*input.size[1]];
	if (regionmap==0)
		return OUT_OF_MEMORY;
	for (unsigned int i=0;i<input.size[0]*input.size[1];i++){
		regionmap[i]=0;
	}

	output.maxLabel = 0;
	output.numCmds = 0;
	output.commands = 0;
	ProgramCommandList* lastCommand = 0;

	if (input.size[0] == 0 || input.size[1] == 0){
		// generte empty program that will terminate immideately.
		output.commands = new ProgramCommandList;
		if (output.commands == 0){
			//TODO: clean up heap!
			return OUT_OF_MEMORY;
		}
		output.commands->next = 0;
		output.commands->command = EXIT;
		output.commands->parameter = 0;
		output.commands->label = 0;
		output.commands->id = 0;
		output.numCmds = 1;
		output.commands->cond[0] = 0;
		output.commands->cond[1] = 0;
		output.commands->cond[2] = 0;
		output.commands->cond[3] = 0;
		
		
	}

	CodelCompileStack* compileStack = new CodelCompileStack;
	if (compileStack == 0){
		// TODO: clean up heap!
		return OUT_OF_MEMORY;
	}
	compileStack->codel[0] = 0;
	compileStack->codel[1] = 0;
	compileStack->d = 0;
	compileStack->prev = 0;

	while (compileStack != 0){

		// Get region that correspondens to the current codel
		Region* reg = regionmap[compileStack->codel[0]+compileStack->codel[1]*input.size[0]];
		if (reg == 0){ // should only occur for the very first element of the program.
			reg = createRegion(compileStack->codel, input, regionmap);
			if (reg == 0){
				//TODO: clean up heap!
				return OUT_OF_MEMORY;
			}
		}
		// Get current program flow direction.
		unsigned char d = compileStack->d;

		// remove current element from codel compile stack
		CodelCompileStack* tmpccs = compileStack;
		compileStack = compileStack->prev;
		delete tmpccs;

		if (reg->command[d]!=0){
			// Code for current region with its corresponding direction has already been generated.
			continue;
		}
		reg->command[d] = new ProgramCommandList;
		if (reg->command[d] == 0){
			//TODO: clean up heap!
			return OUT_OF_MEMORY;
		}
		reg->command[d]->label = 0;

		// process all stored jump origins to reg->command[d] (stored in the list reg->comeFrom[d]) and set them to reg->command[d].	
		// this is neccessary to set all cond[0..3] values that are needed.
		while (reg->comeFrom[d] != 0){
			if (reg->comeFrom[d]->pcl != 0){
				// set it
				*(reg->comeFrom[d]->pcl) = reg->command[d];
				// if any jump origin exists that is not the predecessor in the ProgramCommandList set a label for the current command.
				if (reg->command[d]->label == 0) {
					// check if the current command has to be labeled.
					bool cameFromLast = false;
					for (unsigned char i=0;i<4;i++)
						if (reg->comeFrom[d]->pcl == lastCommand->cond+i)
							cameFromLast = true;
					if (!cameFromLast)
						reg->command[d]->label = ++(output.maxLabel);
				}	
			}
			ListOfProgramCommandList* tmp = reg->comeFrom[d];
			reg->comeFrom[d] = reg->comeFrom[d]->next;
			delete tmp;
		}

		// Add generated command into command list
		if (lastCommand == 0 || output.commands == 0){
			output.commands = reg->command[d];
		}else{
			lastCommand->next = reg->command[d];
		}
		lastCommand = reg->command[d];
		lastCommand->next = 0;
		// The command will be determined later after the next codel in the program flow has been found.
		lastCommand->command = 0;
		lastCommand->parameter = reg->size;
		lastCommand->id = output.numCmds++;
		lastCommand->cond[0] = 0;
		lastCommand->cond[1] = 0;
		lastCommand->cond[2] = 0;
		lastCommand->cond[3] = 0;

		// Prepare for the trace loop that finds the codel/region and DP/CC that follows the current region in program flow.
		// After that lastCommand->command can be determined.

		unsigned int nextCodel[2];
		// if true the value in nextCodel should be ignored.
		bool exit = false;
		// save if cc has been toggled in the last step
		bool cc = false;
		// Tracelist to detect loops (especially while going through white regions) and create an exit command in this case.
		struct TraceList {
			unsigned int codel[2];
			unsigned char d;
			TraceList* next;
		} traceList;
		traceList.codel[0] = -1;
		traceList.codel[1] = -1;
		traceList.d = -1;
		traceList.next = 0;
		TraceList* iterator;
		TraceList* last;
		Region* traceRegion = reg;
		// trace loop: find net codel (and DP/CC) in program flow.
		// the next codel position will be stored in nextCodel. the new DP/CC will be stored in d.
		// if no codel exists that could be entered exit will be set to true.
		while (1){
			// get coordinates of a candidate for the next codel in program flow.
			getNeightbourPixel(traceRegion->exitpoints[d],d/2,nextCodel);
			// check if codel is inside the image boundaries.
			if (nextCodel[0]<input.size[0] && nextCodel[1] < input.size[1]){

				// get region corresponding to the codel
				Region* nextRegion = regionmap[nextCodel[0]+nextCodel[1]*input.size[0]];
				if (nextRegion == 0){
					nextRegion = createRegion(nextCodel, input, regionmap);
					if (nextRegion == 0){
						//TODO: clean up heap!
						return OUT_OF_MEMORY;
					}
				}
				// if region color is not black it can be entered. Then the next codel in program flow has been found.
				if (nextRegion->color != BLACK){
					break;
				}
			}
			// for white pixels: change the region to the last pixel in the current direction
			// for non-white pixels: do nothing
			Region* nextRegion = regionmap[traceRegion->exitpoints[d][0]+traceRegion->exitpoints[d][1]*input.size[0]];
			if (nextRegion == 0){
				nextRegion = createRegion(traceRegion->exitpoints[d], input, regionmap);
				if (nextRegion == 0){
					//TODO: clean up heap!
					return OUT_OF_MEMORY;
				}
				if (nextRegion->color != WHITE){
					// a new region has been created at the exit point but this must not happen for non-white regions:
					// exitpoint is a part of the current non-white region
					// TODO: clean up heap!
					return INTERNAL_ERROR;
				}
			}
			traceRegion = nextRegion;
			// Check, if nextCodel und DP/CC already in list. If yes: there is no other region that can be reached
			// from the current region of codels with same color. => generate EXIT command in ProgamCommandList.
			// iterate through traceList
			iterator = &traceList;
			last = iterator;
			while (iterator != 0){
				if (iterator->codel[0] == traceRegion->exitpoints[d][0] && iterator->codel[1] == traceRegion->exitpoints[d][1] && iterator->d == d){
					// Loop detected. EXIT command will be generated later.
					exit = true;
					break;
				}
				last = iterator;
				iterator = iterator->next;
			}
			if (exit){
				// there is no valid nextCodel.
				// EXIT command will be generated later.
				break;
			}
			iterator = new TraceList;
			if (iterator == 0){
				// TODO: clean up heap
				return OUT_OF_MEMORY;
			}
			iterator->next = 0;
			iterator->codel[0] = traceRegion->exitpoints[d][0];
			iterator->codel[1] = traceRegion->exitpoints[d][1];
			iterator->d = d;
			last->next = iterator;

			// get next direction that should be tried
			if (!cc){
				d = (d&0x6) + (((d&0x1)+1)&0x1);
			}else{
				d = (((d&0x6)+2)&0x6) + (d&0x1);
			}
			cc = !cc;
		}

		// clean up traceList in heap.
		iterator = &traceList;
		last = iterator;
		while (iterator != 0){
			iterator = iterator->next;
			if (last != &traceList){
				delete last;
			}
			last = iterator;
		}

		// determine lastCommand->command.

		// get color change (piet command)
		// change = 18: EXIT command
		unsigned char change = EXIT;

		// check if a codel has been found that can be entered next
		if (!exit) {
			// check that nextCodel is valid. otherwise an internal error occured.
			if (!(nextCodel[0] < input.size[0] && nextCodel[1] < input.size[1])){
				// TODO: clean up heap!
				return INTERNAL_ERROR;
			}
			Region* nextRegion = regionmap[nextCodel[0]+nextCodel[1]*input.size[0]];
			if (nextRegion == 0){
				// cannot occur because the region must have been created in the trace loop above to check that the
				// color of the codel is not black
				// TODO: clean up heap!
				return INTERNAL_ERROR;
			}
			unsigned char codel1 = reg->color;
			unsigned char codel2 = nextRegion->color;

			// change = 0: NOP command
			change = NOP;

			if (codel1 < 18 && codel2 < 18){
				unsigned char hueChange = (6 + (codel2%6) - (codel1 % 6))%6;
				unsigned char lightnessChange = (3 + (codel2/6) - (codel1/6))%3;
				// change 1..17: command
				change = lightnessChange+3*hueChange;
			}
		}

		// TODO: check if there are more commands needed to represent the current command: IN/OUT cmds may print a "? " or a " ".
		if ((change == INN || change == INC) && printQuestionMarkOnIn){
			// print "? "
		}

		// save command.
		lastCommand->command = change;

		// If a number has been printed a space may be added...
		if (change == OUTN && spaceAfterNumOut){
			// print " "

			// PUSH ' '
			ProgramCommandList* newCmd = new ProgramCommandList;
			if (newCmd == 0){
				//TODO: clean up heap!
				return OUT_OF_MEMORY;
			}
			newCmd->label = 0;
			newCmd->next = 0;
			newCmd->command = PUSH;
			newCmd->parameter = ' ';
			newCmd->label = 0;
			newCmd->id = (output.numCmds++);
			newCmd->cond[0] = 0;
			newCmd->cond[1] = 0;
			newCmd->cond[2] = 0;
			newCmd->cond[3] = 0;
			lastCommand->cond[0] = newCmd;
			lastCommand->next = newCmd;
			lastCommand = newCmd;

			// OUTC
			newCmd = new ProgramCommandList;
			if (newCmd == 0){
				//TODO: clean up heap!
				return OUT_OF_MEMORY;
			}
			newCmd->label = 0;
			newCmd->next = 0;
			newCmd->command = OUTC;
			newCmd->parameter = 1;
			newCmd->label = 0;
			newCmd->id = (output.numCmds++);
			newCmd->cond[0] = 0;
			newCmd->cond[1] = 0;
			newCmd->cond[2] = 0;
			newCmd->cond[3] = 0;
			lastCommand->cond[0] = newCmd;
			lastCommand->next = newCmd;
			lastCommand = newCmd;
		}

		// set lastCommand->cond[0..3] if possible.
		// otherwise add them to the list of the region that has to be processed to set them.
		if (lastCommand->command != EXIT){
			// get cond[0..3] JUMP-DIRECTIONS!
			Region* nextRegion = regionmap[nextCodel[0]+nextCodel[1]*input.size[0]];
			if (nextRegion == 0){
				//if command is not EXIT a next region must have been found.
				// TODO: clean up heap!
				return INTERNAL_ERROR;
			}
			// get the corresponding d-pointer and find nextRegion->command[d+X].
			unsigned char numCondSet = 1;
			unsigned char condDirections[4];
			condDirections[0] = d;
			if (lastCommand->command == SWTCH){
				numCondSet = 2;
				condDirections[1] = (((d&0x1)+1)&0x1)+(d&0x6);
			} else if (lastCommand->command == PNTR){
				numCondSet = 4;
				for (unsigned char i=1;i<4;i++)
					condDirections[i] = (d&0x1)+(((d&0x6)+2*(i))&0x6);
			}
			// if nextRegion->command[d+X] is 0 then add lastCommand->condX to nextRegion->comeFrom[d+X] list and add nextRegion+d to compileStack.
			// otherwise add a label to the nextRegion->command[d+X] if not already set.
			for (unsigned char condition=0;condition<numCondSet;condition++){
				if (nextRegion->command[condDirections[condition]] != 0){
					if (nextRegion->command[condDirections[condition]]->label == 0){
						nextRegion->command[condDirections[condition]]->label = ++(output.maxLabel);
					}
					// add nextRegion->command[condDirections[condition]] to lastCommand->cond[condition]
					lastCommand->cond[condition] = nextRegion->command[condDirections[condition]];
				}else{
					// add to comeFrom list.
					ListOfProgramCommandList* newElement = new ListOfProgramCommandList;
					if (newElement==0){
						//TODO: clean up heap!
						return OUT_OF_MEMORY;
					}
					newElement->next = nextRegion->comeFrom[condDirections[condition]];
					newElement->pcl = lastCommand->cond + condition;
					nextRegion->comeFrom[condDirections[condition]] = newElement;

					// add to compileStack!
					CodelCompileStack* newCompileStackElement = new CodelCompileStack;
					if (newCompileStackElement==0){
						//TODO: clean up heap!
						return OUT_OF_MEMORY;
					}
					newCompileStackElement->codel[0] = nextCodel[0];
					newCompileStackElement->codel[1] = nextCodel[1];
					newCompileStackElement->d = condDirections[condition];
					newCompileStackElement->prev = compileStack;
					compileStack = newCompileStackElement;
				}
			}
		}
	}

	// MAYBE TODO: remove NOPs.

	// TODO: change labels to be ordered.

	return SUCCESS;
}

Region* createRegion(unsigned int codel[2], PietImage image, Region** regionmap){

	if (regionmap == 0)
		return 0;
	if (codel[0] >= image.size[0] || codel[1] >= image.size[1])
		return 0;

	if (regionmap[codel[0]+codel[1]*image.size[0]] != 0)
		return regionmap[codel[0]+codel[1]*image.size[0]];

	Region* region = new Region;
	if (region == 0)
		return 0;

	region->color = image.image[codel[0]+codel[1]*image.size[0]];

	region->size = 1;
	regionmap[codel[0]+codel[1]*image.size[0]] = region;

	if (region->color == WHITE){
		// region has size of 1, but: exitpoint may lie outside the region,
		// it is the last point that is WHITE while going in the same direction

		// determine exitpoints
		unsigned int curPos[2];
		unsigned int newPos[2];
		for (unsigned char i=0;i<4;i++){
			newPos[0] = codel[0];
			newPos[1] = codel[1];
			do{
				curPos[0] = newPos[0];
				curPos[1] = newPos[1];
				getNeightbourPixel(curPos,i,newPos);
				if (newPos[0] >= image.size[0] || newPos[1] >= image.size[1])
					break;
			}while(image.image[newPos[0]+newPos[1]*image.size[0]] == WHITE);
			region->exitpoints[2*i][0] = curPos[0];
			region->exitpoints[2*i][1] = curPos[1];
			region->exitpoints[2*i+1][0] = curPos[0];
			region->exitpoints[2*i+1][1] = curPos[1];
			region->command[2*i] = 0;
			region->command[2*i+1] = 0;
			region->comeFrom[2*i] = 0;
			region->comeFrom[2*i+1] = 0;
		}
		return region;
	}

	// initialize exit points with current pixel position
	for (unsigned char i=0;i<8;i++){
		for (int j=0;j<2;j++)
			region->exitpoints[i][j] = codel[j];
		region->command[i] = 0;
		region->comeFrom[i] = 0;
	}

	// now handle regions bigger than 1 pixel: each pixel that borders to the current region and has the same color will be added to the region iterative.
	// to store pixels that have to be added use a stack.

	PixelStack* pixelStack = new PixelStack;
	PixelStack* old = 0;
	pixelStack->previous = 0;
	pixelStack->position[0] = codel[0];
	pixelStack->position[1] = codel[1];

	unsigned int curPos[2];
	unsigned int newPos[2];

	while (pixelStack != 0){

		curPos[0] = pixelStack->position[0];
		curPos[1] = pixelStack->position[1];
		old = pixelStack;
		pixelStack = pixelStack->previous;
		delete old;
		old = 0;
		for (unsigned char i=0;i<4;i++){
			getNeightbourPixel(curPos, i, newPos);
			if (newPos[0] < image.size[0] && newPos[1] < image.size[1]){
				if (image.image[newPos[0]+newPos[1]*image.size[0]] == region->color){
					// test if a region for this pixel already exists (TODO: if it is anozther region than the current treated this is an internal error!)
					// otherwise add pixel to the current region.
					if (regionmap[newPos[0]+newPos[1]*image.size[0]] == 0){
						// add region to regionmap at new pixel position
						regionmap[newPos[0]+newPos[1]*image.size[0]] = region;
						region->size++;
						//refresh EXIT-POINTS:
						if (newPos[0] > region->exitpoints[0][0] || (newPos[0] == region->exitpoints[0][0] && newPos[1] < region->exitpoints[0][1])){
							region->exitpoints[0][0] = newPos[0];
							region->exitpoints[0][1] = newPos[1];
						}
						if (newPos[0] > region->exitpoints[1][0] || (newPos[0] == region->exitpoints[1][0] && newPos[1] > region->exitpoints[1][1])){
							region->exitpoints[1][0] = newPos[0];
							region->exitpoints[1][1] = newPos[1];
						}
						if (newPos[1] > region->exitpoints[2][1] || (newPos[1] == region->exitpoints[2][1] && newPos[0] > region->exitpoints[2][0])){
							region->exitpoints[2][0] = newPos[0];
							region->exitpoints[2][1] = newPos[1];
						}
						if (newPos[1] > region->exitpoints[3][1] || (newPos[1] == region->exitpoints[3][1] && newPos[0] < region->exitpoints[3][0])){
							region->exitpoints[3][0] = newPos[0];
							region->exitpoints[3][1] = newPos[1];
						}
						if (newPos[0] < region->exitpoints[4][0] || (newPos[0] == region->exitpoints[4][0] && newPos[1] > region->exitpoints[4][1])){
							region->exitpoints[4][0] = newPos[0];
							region->exitpoints[4][1] = newPos[1];
						}
						if (newPos[0] < region->exitpoints[5][0] || (newPos[0] == region->exitpoints[5][0] && newPos[1] < region->exitpoints[5][1])){
							region->exitpoints[5][0] = newPos[0];
							region->exitpoints[5][1] = newPos[1];
						}
						if (newPos[1] < region->exitpoints[6][1] || (newPos[1] == region->exitpoints[6][1] && newPos[0] < region->exitpoints[6][0])){
							region->exitpoints[6][0] = newPos[0];
							region->exitpoints[6][1] = newPos[1];
						}
						if (newPos[1] < region->exitpoints[7][1] || (newPos[1] == region->exitpoints[7][1] && newPos[0] > region->exitpoints[7][0])){
							region->exitpoints[7][0] = newPos[0];
							region->exitpoints[7][1] = newPos[1];
						}
						//add neightbour pixels to stack
						old = pixelStack;
						pixelStack = new PixelStack;
						pixelStack->previous = old;
						old = 0;
						pixelStack->position[0] = newPos[0];
						pixelStack->position[1] = newPos[1];
					}
				}
			}
		}
	}
	return region;
}

void getNeightbourPixel(unsigned int position[2], unsigned char direction, unsigned int neightbour[2]){
	direction %= 4;
	switch (direction){
		case 0:
			neightbour[0] = position[0]+1;
			neightbour[1] = position[1];
			break;
		case 1:
			neightbour[0] = position[0];
			neightbour[1] = position[1]+1;
			break;
		case 2:
			neightbour[0] = position[0]-1;
			neightbour[1] = position[1];
			break;
		case 3:
			neightbour[0] = position[0];
			neightbour[1] = position[1]-1;
			break;
		default: //should not occur
			neightbour[0] = position[0];
			neightbour[1] = position[1];
			break;
	}
}


